(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TurboPower Abbrevia
 *
 * The Initial Developer of the Original Code is
 * TurboPower Software
 *
 * Portions created by the Initial Developer are Copyright (C) 1997-2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** *)

{**********************************************************************
 Package pl_Abbrevia
 From PilotLogic Software House (http://www.pilotlogic.com/)
 This unit is part of CodeTyphon Studio
***********************************************************************}

unit AbBrowse;

{$I AbDefine.inc}

interface

uses
  Classes,
  Controls, Graphics, Forms, ExtCtrls,
  Types,
  AbConst,
  AbBase,
  AbUtils,
  AbArcTyp;

type
  IAbProgressMeter = interface
    ['{4B766704-FD20-40BF-BA40-2EC2DD77B178}']
    procedure DoProgress(Progress: byte);
    procedure Reset;
  end;


  TAbMeterOrientation = (moHorizontal, moVertical);

  TAbCustomMeter = class(TGraphicControl, IAbProgressMeter)
  protected
    FBorderStyle: TBorderStyle;
    FCtl3D: boolean;
    FOrientation: TAbMeterOrientation;
    FPercent: integer;
    FTickMarks: byte;
    FUsedColor: TColor;
    FUnusedColor: TColor;
    function GetVersion: string;
    procedure Paint; override;
    procedure SetBorderStyle(const Value: TBorderStyle);
    procedure SetCtl3D(const Value: boolean);
    procedure SetOrientation(const O: TAbMeterOrientation);
    procedure SetTickMarks(const Value: byte);
    procedure SetUnusedColor(const C: TColor);
    procedure SetUsedColor(const C: TColor);
    procedure SetVersion(Value: string);
    property Version: string read GetVersion write SetVersion stored False;
  public
    constructor Create(AOwner: TComponent); override;
    procedure DoProgress(Progress: byte);
    procedure Reset;
  public
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property Ctl3D: boolean read FCtl3D write SetCtl3D default True;
    property Orientation: TAbMeterOrientation read FOrientation write SetOrientation;
    property TickMarks: byte read FTickMarks write SetTickMarks default 10;
    property UnusedColor: TColor read FUnusedColor write SetUnusedColor;
    property UsedColor: TColor read FUsedColor write SetUsedColor;
  end;

  TAbCustomMeterClass = class of TAbCustomMeter;


  TAbMeter = class(TAbCustomMeter)
  published
    property Anchors;
    property Constraints;
    property Align;
    property BorderStyle;
    property Ctl3D;
    property Font;
    property OnClick;
    property OnDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property Orientation;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property TickMarks;
    property UnusedColor;
    property UsedColor;
    property Version;
    property Visible;
  end;


  TAbBaseBrowser = class(TAbBaseComponent)
  public
    FArchive: TAbArchive;
  protected
    FSpanningThreshold: longint;
    FItemProgressMeter: TAbCustomMeter;
    FArchiveProgressMeter: TAbCustomMeter;
    FBaseDirectory: string;
    FFileName: string;
    FLogFile: string;
    FLogging: boolean;
    FOnArchiveProgress: TAbArchiveProgressEvent;
    FOnArchiveItemProgress: TAbArchiveItemProgressEvent;
    FOnChange: TNotifyEvent;
    FOnConfirmProcessItem: TAbArchiveItemConfirmEvent;
    FOnLoad: TAbArchiveEvent;
    FOnProcessItemFailure: TAbArchiveItemFailureEvent;
    FOnRequestImage: TAbRequestImageEvent;
    FTempDirectory: string;

    FArchiveType: TAbArchiveType;
    FForceType: boolean;

  protected
    function GetCount: integer;
    function GetItem(Value: longint): TAbArchiveItem;
    function GetSpanned: boolean;
    function GetStatus: TAbArchiveStatus;
    procedure ResetMeters; virtual;
    procedure SetArchiveProgressMeter(const Value: TAbCustomMeter);
    procedure SetCompressionType(const Value: TAbArchiveType);
    procedure SetBaseDirectory(const Value: string);
    procedure SetItemProgressMeter(const Value: TAbCustomMeter);
    procedure SetSpanningThreshold(Value: longint);
    procedure SetLogFile(const Value: string);
    procedure SetLogging(Value: boolean);
    procedure SetTempDirectory(const Value: string);

    procedure Loaded; override;
    procedure Notification(Component: TComponent;
      Operation: TOperation); override;

  protected {virtual methods}
    procedure DoArchiveItemProgress(Sender: TObject;
      Item: TAbArchiveItem;
      Progress: byte;
      var Abort: boolean); virtual;
    procedure DoArchiveProgress(Sender: TObject;
      Progress: byte;
      var Abort: boolean); virtual;
    procedure DoChange; virtual;
    procedure DoConfirmProcessItem(Sender: TObject;
      Item: TAbArchiveItem;
      ProcessType: TAbProcessType;
      var Confirm: boolean); virtual;
    procedure DoLoad(Sender: TObject); virtual;
    procedure DoProcessItemFailure(Sender: TObject;
      Item: TAbArchiveItem;
      ProcessType: TAbProcessType;
      ErrorClass: TAbErrorClass;
      ErrorCode: integer); virtual;
    procedure SetOnRequestImage(Value: TAbRequestImageEvent); virtual;
    procedure InitArchive; virtual;

    {This method must be defined in descendent classes}
    procedure SetFileName(const aFileName: string); virtual; abstract;


  protected {properties}
    property Archive: TAbArchive read FArchive;
    property ArchiveProgressMeter: TAbCustomMeter
      read FArchiveProgressMeter write SetArchiveProgressMeter;
    property BaseDirectory: string read FBaseDirectory write SetBaseDirectory;
    property FileName: string read FFileName write SetFileName;
    property SpanningThreshold: longint read FSpanningThreshold
      write SetSpanningThreshold default 0;
    property ItemProgressMeter: TAbCustomMeter
      read FItemProgressMeter write SetItemProgressMeter;
    property LogFile: string read FLogFile write SetLogFile;
    property Logging: boolean read FLogging write SetLogging default False;
    property Spanned: boolean read GetSpanned;
    property TempDirectory: string read FTempDirectory write SetTempDirectory;



  protected {events}
    property OnArchiveProgress: TAbArchiveProgressEvent
      read FOnArchiveProgress write FOnArchiveProgress;
    property OnArchiveItemProgress: TAbArchiveItemProgressEvent
      read FOnArchiveItemProgress write FOnArchiveItemProgress;
    property OnConfirmProcessItem: TAbArchiveItemConfirmEvent
      read FOnConfirmProcessItem write FOnConfirmProcessItem;
    property OnProcessItemFailure: TAbArchiveItemFailureEvent
      read FOnProcessItemFailure write FOnProcessItemFailure;
    property OnRequestImage: TAbRequestImageEvent
      read FOnRequestImage write SetOnRequestImage;

  public {methods}
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ClearTags;
    {Clear all tags from the archive}
    function FindItem(aItem: TAbArchiveItem): integer;
    function FindFile(const aFileName: string): integer;
    procedure TagItems(const FileMask: string);
    {tag all items that match the mask}
    procedure UnTagItems(const FileMask: string);
    {clear tags for all items that match the mask}
    procedure CloseArchive;
    {closes the archive by setting FileName to ''}
    procedure OpenArchive(const aFileName: string);
    {opens the archive}

  public {properties}
    property Count: integer read GetCount;
    property Items[Index: integer]: TAbArchiveItem read GetItem; default;
    property Status: TAbArchiveStatus read GetStatus;

    property ArchiveType: TAbArchiveType read FArchiveType
      write SetCompressionType default atUnknown;

    property ForceType: boolean read FForceType write FForceType default False;

  public {events}
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnLoad: TAbArchiveEvent read FOnLoad write FOnLoad;
  end;

function AbDetermineArcType(const FN: string;
  AssertType: TAbArchiveType): TAbArchiveType; overload;
function AbDetermineArcType(aStream: TStream): TAbArchiveType; overload;

implementation

uses
  SysUtils,
  AbExcept,
{$IFDEF MSWINDOWS}
  AbCabTyp,
{$ENDIF}
  AbZipTyp,
  AbTarTyp,
  AbGzTyp,
  AbBzip2Typ;

{ TAbBaseBrowser implementation ======================================= }

{ -------------------------------------------------------------------------- }
constructor TAbBaseBrowser.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLogFile := '';
  FLogging := False;
  FSpanningThreshold := 0;

  FArchiveType := atUnknown;
  FForceType := False;
end;

{ -------------------------------------------------------------------------- }
destructor TAbBaseBrowser.Destroy;
begin
  FArchive.Free;
  FArchive := nil;
  inherited Destroy;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.ClearTags;
{Clear all tags from the archive}
begin
  if Assigned(FArchive) then
    FArchive.ClearTags
  else
    raise EAbNoArchive.Create;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.CloseArchive;
{closes the archive by setting FileName to ''}
begin
  if FFileName <> '' then
    FileName := '';
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.DoArchiveItemProgress(Sender: TObject;
  Item: TAbArchiveItem;
  Progress: byte;
  var Abort: boolean);
begin
  Abort := False;
  if Assigned(FItemProgressMeter) then
    FItemProgressMeter.DoProgress(Progress);
  if Assigned(FOnArchiveItemProgress) then
    FOnArchiveItemProgress(Self, Item, Progress, Abort);
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.DoArchiveProgress(Sender: TObject;
  Progress: byte;
  var Abort: boolean);
begin
  Abort := False;
  if Assigned(FArchiveProgressMeter) then
    FArchiveProgressMeter.DoProgress(Progress);
  if Assigned(FOnArchiveProgress) then
    FOnArchiveProgress(Self, Progress, Abort);
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.DoChange;
begin
  if Assigned(FOnChange) then
  begin
    FOnChange(Self);
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.DoConfirmProcessItem(Sender: TObject;
  Item: TAbArchiveItem;
  ProcessType: TAbProcessType;
  var Confirm: boolean);
begin
  Confirm := True;
  if Assigned(FItemProgressMeter) then
    FItemProgressMeter.Reset;
  if Assigned(FOnConfirmProcessItem) then
    FOnConfirmProcessItem(Self, Item, ProcessType, Confirm);
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.DoLoad(Sender: TObject);
begin
  if Assigned(FOnLoad) then
    FOnLoad(Self);
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.DoProcessItemFailure(Sender: TObject;
  Item: TAbArchiveItem;
  ProcessType: TAbProcessType;
  ErrorClass: TAbErrorClass;
  ErrorCode: integer);
begin
  if Assigned(FOnProcessItemFailure) then
    FOnProcessItemFailure(Self, Item, ProcessType, ErrorClass, ErrorCode);
end;

{ -------------------------------------------------------------------------- }
function TAbBaseBrowser.FindItem(aItem: TAbArchiveItem): integer;
begin
  if Assigned(FArchive) then
    Result := FArchive.FindItem(aItem)
  else
    Result := -1;
end;

{ -------------------------------------------------------------------------- }
function TAbBaseBrowser.FindFile(const aFileName: string): integer;
begin
  if Assigned(FArchive) then
    Result := FArchive.FindFile(aFileName)
  else
    Result := -1;
end;

{ -------------------------------------------------------------------------- }
function TAbBaseBrowser.GetSpanned: boolean;
begin
  if Assigned(FArchive) then
    Result := FArchive.Spanned
  else
    Result := False;
end;

{ -------------------------------------------------------------------------- }
function TAbBaseBrowser.GetStatus: TAbArchiveStatus;
begin
  if Assigned(FArchive) then
    Result := FArchive.Status
  else
    Result := asInvalid;
end;

{ -------------------------------------------------------------------------- }
function TAbBaseBrowser.GetCount: integer;
begin
  if Assigned(FArchive) then
    Result := FArchive.Count
  else
    Result := 0;
end;

{ -------------------------------------------------------------------------- }
function TAbBaseBrowser.GetItem(Value: longint): TAbArchiveItem;
begin
  if Assigned(FArchive) then
    Result := FArchive.ItemList[Value]
  else
    raise EAbNoArchive.Create;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.InitArchive;
begin
  ResetMeters;
  if Assigned(FArchive) then
  begin
    {properties}
    FArchive.SpanningThreshold := FSpanningThreshold;
    FArchive.LogFile := FLogFile;
    FArchive.Logging := FLogging;
    FArchive.TempDirectory := FTempDirectory;
    SetBaseDirectory(FBaseDirectory);
    {events}
    FArchive.OnArchiveProgress := DoArchiveProgress;
    FArchive.OnArchiveItemProgress := DoArchiveItemProgress;
    FArchive.OnConfirmProcessItem := DoConfirmProcessItem;
    FArchive.OnLoad := DoLoad;
    FArchive.OnProcessItemFailure := DoProcessItemFailure;
    FArchive.OnRequestImage := FOnRequestImage;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.Loaded;
begin
  inherited Loaded;
  DoChange;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.Notification(Component: TComponent;
  Operation: TOperation);
begin
  inherited Notification(Component, Operation);
  if (Operation = opRemove) then
  begin
    if Assigned(ItemProgressMeter) and Component.IsImplementorOf(ItemProgressMeter) then
      ItemProgressMeter := nil;
    if Assigned(ArchiveProgressMeter) and Component.IsImplementorOf(
      ArchiveProgressMeter) then
      ArchiveProgressMeter := nil;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.OpenArchive(const aFileName: string);
{opens the archive}
begin
  FileName := AFileName;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.ResetMeters;
begin
  if Assigned(FArchiveProgressMeter) then
    FArchiveProgressMeter.Reset;
  if Assigned(FItemProgressMeter) then
    FItemProgressMeter.Reset;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetBaseDirectory(const Value: string);
begin
  if Assigned(FArchive) then
  begin
    FArchive.BaseDirectory := Value;
    FBaseDirectory := FArchive.BaseDirectory;
  end
  else
    FBaseDirectory := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetSpanningThreshold(Value: longint);
begin
  FSpanningThreshold := Value;
  if Assigned(FArchive) then
    FArchive.SpanningThreshold := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetLogFile(const Value: string);
begin
  FLogFile := Value;
  if (csDesigning in ComponentState) then
    Exit;
  if Assigned(FArchive) then
    FArchive.LogFile := Value;
  SetLogging(Value <> '');
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetLogging(Value: boolean);
begin
  FLogging := Value;
  if (csDesigning in ComponentState) then
    Exit;
  if Assigned(FArchive) then
    FArchive.Logging := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetOnRequestImage(Value: TAbRequestImageEvent);
begin
  FOnRequestImage := Value;
  if Assigned(FArchive) then
    FArchive.OnRequestImage := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetTempDirectory(const Value: string);
begin
  FTempDirectory := Value;
  if Assigned(FArchive) then
    FArchive.TempDirectory := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.TagItems(const FileMask: string);
{tag all items that match the mask}
begin
  if Assigned(FArchive) then
    FArchive.TagItems(FileMask)
  else
    raise EAbNoArchive.Create;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.UnTagItems(const FileMask: string);
{clear tags for all items that match the mask}
begin
  if Assigned(FArchive) then
    FArchive.UnTagItems(FileMask)
  else
    raise EAbNoArchive.Create;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetCompressionType(const Value: TAbArchiveType);
begin
  if not Assigned(FArchive) or (Status <> asInvalid) then
    FArchiveType := Value
  else
    raise EAbArchiveBusy.Create;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetArchiveProgressMeter(const Value: TAbCustomMeter);
begin
  FArchiveProgressMeter := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbBaseBrowser.SetItemProgressMeter(const Value: TAbCustomMeter);
begin
  FItemProgressMeter := Value;
end;

{ -------------------------------------------------------------------------- }
function AbDetermineArcType(const FN: string;
  AssertType: TAbArchiveType): TAbArchiveType;
var
  Ext: string;
  FS: TFileStream;
begin
  Result := AssertType;
  if Result = atUnknown then
  begin
    { Guess archive type based on it's extension }
    Ext := UpperCase(ExtractFileExt(FN));
    if (Ext = '.ZIP') or (Ext = '.JAR') then
      Result := atZip;
    if (Ext = '.EXE') then
      Result := atSelfExtZip;
    if (Ext = '.TAR') then
      Result := atTar;
    if (Ext = '.GZ') then
      Result := atGzip;
    if (Ext = '.TGZ') then
      Result := atGzippedTar;
    if (Ext = '.CAB') then
      Result := atCab;
    if (Ext = '.BZ2') then
      Result := atBzip2;
    if (Ext = '.TBZ') then
      Result := atBzippedTar;
  end;
  {$IFNDEF MSWINDOWS}
  if Result = atCab then
    Result := atUnknown;
  {$ENDIF}
  if FileExists(FN) and (AbFileGetSize(FN) > 0) then
  begin
    { If the file doesn't exist (or is empty) presume to make one, otherwise
      guess or verify the contents }
    FS := TFileStream.Create(FN, fmOpenRead or fmShareDenyNone);
    try
      if Result = atUnknown then
        Result := AbDetermineArcType(FS)
      else
      begin
        case Result of
          atZip:
          begin
            Result := VerifyZip(FS);
          end;
          atSelfExtZip:
          begin
            Result := VerifySelfExtracting(FS);
          end;
          atTar:
          begin
            Result := VerifyTar(FS);
          end;
          atGzip, atGzippedTar:
          begin
            Result := VerifyGzip(FS);
          end;
          {$IFDEF MSWINDOWS}
          atCab:
          begin
            Result := VerifyCab(FS);
          end;
          {$ENDIF}
          atBzip2, atBzippedTar:
          begin
            Result := VerifyBzip2(FS);
          end;
        end;
      end;
    finally
      FS.Free;
    end;
  end;
end;

{ -------------------------------------------------------------------------- }
function AbDetermineArcType(aStream: TStream): TAbArchiveType;
begin
  { VerifyZip returns true for self-extracting zips too, so test those first }
  Result := VerifySelfExtracting(aStream);
  if Result = atUnknown then
    Result := VerifyZip(aStream);
  if Result = atUnknown then
    Result := VerifyTar(aStream);
  if Result = atUnknown then
    Result := VerifyGzip(aStream);
  if Result = atUnknown then
    Result := VerifyBzip2(aStream);
  {$IFDEF MSWINDOWS}
  if Result = atUnknown then
    Result := VerifyCab(aStream);
  {$ENDIF}
end;

{ -------------------------------------------------------------------------- }

{ == TAbCustomMeter ======================================================== }
constructor TAbCustomMeter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFNDEF UsingCLX}
  if NewStyleControls then
    ControlStyle := ControlStyle + [csOpaque]
  else
    ControlStyle := ControlStyle + [csOpaque, csFramed];
  {$ELSE}
  ControlStyle := ControlStyle + [csOpaque, csFramed];
  {$ENDIF}

  FBorderStyle := bsSingle;
  FCtl3D := True;
  FOrientation := moHorizontal;
  FTickMarks := 10;
  FUnusedColor := clBtnFace;
  FUsedColor := clNavy;
  Width := 150;
  Height := 16;
end;

{ -------------------------------------------------------------------------- }
function TAbCustomMeter.GetVersion: string;
begin
  Result := AbVersionS;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.DoProgress(Progress: byte);
begin
  if (Progress <> FPercent) then
  begin
    FPercent := Progress;
    if (FPercent >= 100) then
      FPercent := 0;
    Refresh;
    Application.ProcessMessages;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.Paint;
const
  VSpace = 2;
  HSpace = 1;
  LSpace = 1;
  RSpace = 1;
var
  ClRect, R: TRect;
  ClWidth: integer;
  ClHeight: integer;
  BlockWidth: integer;
  BlockCount: integer;
  i: integer;
begin
  ClRect := ClientRect;

  ClWidth := ClRect.Right - CLRect.Left + 1;
  ClHeight := ClRect.Bottom - ClRect.Top + 1;

  if (Orientation = moHorizontal) then
    BlockWidth := ((ClWidth - LSpace - RSpace - (9 * VSpace)) div FTickMarks) + 1
  else
    BlockWidth := ((ClHeight - LSpace - RSpace - (9 * HSpace)) div FTickMarks) + 1;
  BlockCount := FPercent div FTickMarks;

  if not Assigned((Canvas as TControlCanvas).Control) then
  begin
    TControlCanvas(Canvas).Control := self;
  end;
  with Canvas do
  begin
    Brush.Color := FUnusedColor;
    FillRect(Rect(ClRect.Left, ClRect.Top, ClRect.Left + ClWidth -
      1, ClRect.Top + ClHeight - 1));
    Brush.Color := FUsedColor;
    if (BlockCount > 0) then
    begin
      if (Orientation = moHorizontal) then
      begin
        R.Top := ClRect.Top + HSpace;
        R.Bottom := ClRect.Bottom - HSpace;
        for i := 0 to Pred(BlockCount) do
        begin
          R.Left := ClRect.Left + LSpace + (i * VSpace) + (i * BlockWidth);
          R.Right := R.Left + BlockWidth;
          FillRect(R);
        end;
      end
      else
      begin {moVertical}
        R.Left := ClRect.Left + VSpace;
        R.Right := ClRect.Right - VSpace;
        for i := 0 to Pred(BlockCount) do
        begin
          R.Bottom := ClRect.Bottom - LSpace - (i * HSpace) -
            (i * BlockWidth);
          R.Top := R.Bottom - BlockWidth;
          FillRect(R);
        end;
      end;
    end;
  end;
  {$IFNDEF LCL}
  if (BorderStyle <> bsNone) then
  begin
    if Ctl3D then
      Frame3D(Canvas, ClRect, clBtnShadow, clBtnHighlight, 1)
    else
      Frame3D(Canvas, ClRect, clBlack, clBlack, 1);
  end;
  {$ENDIF}
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.Reset;
begin
  DoProgress(0);
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetBorderStyle(const Value: TBorderStyle);
begin
  if (Value <> FBorderStyle) then
  begin
    FBorderStyle := Value;
    Invalidate;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetCtl3D(const Value: boolean);
begin
  if (Value <> FCtl3D) then
  begin
    FCtl3D := Value;
    Invalidate;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetOrientation(const O: TAbMeterOrientation);
var
  Temp: integer;
begin
  if (O <> FOrientation) then
  begin
    FOrientation := O;
    if not (csLoading in ComponentState) then
    begin
      Temp := Width;
      Width := Height;
      Height := Temp;
    end;
    Invalidate;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetTickMarks(const Value: byte);
begin
  if Value <= 0 then
    FTickMarks := 10
  else
    FTickMarks := Value;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetUnusedColor(const C: TColor);
begin
  if (C <> FUnusedColor) then
  begin
    FUnusedColor := C;
    Invalidate;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetUsedColor(const C: TColor);
begin
  if (C <> FUsedColor) then
  begin
    FUsedColor := C;
    Invalidate;
  end;
end;

{ -------------------------------------------------------------------------- }
procedure TAbCustomMeter.SetVersion(Value: string);
begin
  {NOP}
end;

initialization

  registerclass(TAbCustomMeter);

end.
